from shutil import copy2

from .Step import Step
from autoinsight.common.Utils import makeDirs


class Copy(Step):
    def __init__(self, source: str, destination: str, isOverWrite: bool = False):
        super().__init__(Copy.__name__, Copy.__name__, self.os, self.os)
        self._source: str = source
        self._destination: str = destination
        self._isOverWrite: bool = isOverWrite

    def __repr__(self) -> str:
        return f"step {self.name} from {self._source} to {self._destination}"

    def __str__(self) -> str:
        return self.__repr__()

    def execute(self, isPaused: bool = False):
        super().execute(isPaused)
        makeDirs(self._destination, self._isOverWrite)
        copy2(self._source, self._destination)



import os
import re
import shutil
import logging
import platform

from enum import Enum
from difflib import get_close_matches, SequenceMatcher
from typing import Collection, Tuple, Iterable, Callable, Any, Optional, Union
from uuid import uuid4, UUID
from pathlib import Path
from time import strftime, gmtime, time_ns

from .EnumTypes import BinaryTypes, OSTypes


def GUID() -> UUID:
    return uuid4()


def strGUID() -> str:
    return str(uuid4())


def strToGUID(strGuid: str):
    return UUID(f"{strGuid}")


# TODO refactor the method
def toUniqueList(items: Iterable[str]) -> Collection[str]:
    uniqueList = []
    illegal = -1
    patterns = [r"\d", "[a-z]", "[A-Z]", r"\W", r"[^\d|a-z|A-Z|\s|\W]"]

    for item in items:
        if isinstance(item, str) and not item.strip():
            continue

        words = re.split(r"\s", item)
        for word in words:
            previous = None
            splitIndexes = []
            for i, c in enumerate(word):
                for p in patterns:
                    if re.search(p, c) and previous != p:
                        splitIndexes.append(i)
                        previous = p
                        break

            length = len(splitIndexes)
            for i, v in enumerate(splitIndexes):
                j = i + 1
                if j < length:
                    if v + 1 == splitIndexes[j]:
                        splitIndexes[j] = illegal

            if length - splitIndexes.count(illegal) < 2:
                frag = word.lower()
                if frag not in uniqueList:
                    uniqueList.append(frag)

                continue

            i = 0
            for j in splitIndexes[1:]:
                if j == illegal:
                    continue
                frag = word[i:j].lower()
                if frag not in uniqueList:
                    uniqueList.append(frag)
                i = j

            frag = word[i:].lower()
            if frag not in uniqueList:
                uniqueList.append(frag)

    return uniqueList


def toAllPossibleGroups(items: Iterable[str]) -> Collection[str]:
    possibles = []
    for i in range(1, len(items) + 1):
        possibles.append(' '.join(items[:i]))
    return possibles


def matchScore(query: str, descriptions: Iterable[str]) -> Tuple[float, float]:
    if not (descriptions and query):
        return 0.0, 0.0

    unique = toUniqueList(descriptions)
    if unique:
        words = toUniqueList(re.split(r"\s", query))
        possibles = toAllPossibleGroups(words)
        matchCount = 0
        possibleCount = 0
        for word in words:
            matches = get_close_matches(word, unique)
            matchCount += len(matches)

        for possible in possibles:
            matches = get_close_matches(possible, unique)
            possibleCount += len(matches)

        lengthUnique = len(unique)
        lengthWord = len(words)
        maxLength = max(lengthUnique, lengthWord)
        firstScore = possibleCount / lengthWord
        secondScore = (firstScore + matchCount) / maxLength
        return firstScore, secondScore
    else:
        return 0.0, 0.0


def isComparable(obj: Any) -> bool:
    try:
        obj < obj
        return True
    except:
        return False


def first(collection: Iterable[Any], isRevert: bool = False, filterFunc: Callable[Any, bool] = None, sortKeyFunc: Optional[Callable[Any, int]] = None, isSort: bool = True) -> Any:
    try:
        if not collection:
            return

        collection = tuple(collection)
        isSort = isComparable(collection[0]) and isSort
        if not (isSort or sortKeyFunc):
            return next(filter(filterFunc, collection))
        else:
            return next(filter(filterFunc, sorted(collection, key=sortKeyFunc, reverse=isRevert)))
    except StopIteration:
        return


def isIEqual(str0: str, str1: str) -> bool:
    """
    Test two strings are ignored case equaled
    """
    if isinstance(str1, str) and isinstance(str0, str):
        return str0.lower() == str1.lower()
    else:
        return False


def isSimilar(str0: Union[str, Enum], str1: Union[str, Enum]) -> bool:
    """
    Check two strings are similar or not

    Args:
        str0 (str): First string
        str1 (str): Second string

    Returns:
        bool: return True if the two strings contain similar contents.
    """
    return SequenceMatcher(None, str(str0).lower(), str(str1).lower()).ratio() > 0.2


def similarRate(str0: Union[str, Enum], str1: Union[str, Enum]) -> float:
    """
    Check two strings are similar or not

    Args:
        str0 (str): First string
        str1 (str): Second string

    Returns:
        bool: return True if the two strings contain similar contents.
    """
    return SequenceMatcher(None, str(str0).lower(), str(str1).lower()).ratio()


def makeDirs(path: str, isRecreate: bool = False):
    isExist = os.path.isdir(path)
    if isRecreate and isExist:
        try:
            shutil.rmtree(path)
        except Exception as e:
            logging.warning("Call makeDirs with ", path, isRecreate, " failed with error", e)
            pass

    elif isExist:
        return

    else:
        os.makedirs(path)


def decorateFileName(name: str) -> str:
    p = Path(name)

    if p.parent and str(p.parent) != ".":
        parent = p.parent
    else:
        parent = ""

    return f'{parent}{strftime("UTC%Y%m%d-%H%M%S", gmtime())}.{str(time_ns())[-9:-6]}-{p.name}'


def getOSType() -> OSTypes:
    platform = platform.system()
    if platform == OSTypes.Windows.name:
        return OSTypes.Windows
    elif platform == OSTypes.MacOS.name:
        return OSTypes.MacOS
    elif platform == OSTypes.Linux.name:
        return OSTypes.Linux
    elif platform == OSTypes.Java.name:
        return OSTypes.Java
    else:
        return OSTypes.Any


def getExecutableBinaryTypes(path: Path) -> bool:
    if getOSType() == OSTypes.Windows:
        import win32file

        if path.exists():
            binaryType = win32file.GetBinaryType(path)
            if binaryType == win32file.SCS_32BIT_BINARY:
                return BinaryTypes.Bit32
            elif binaryType == win32file.SCS_DOS_BINARY         \
                    or binaryType == win32file.SCS_WOW_BINARY   \
                    or binaryType == win32file.SCS_PIF_BINARY   \
                    or binaryType == win32file.SCS_OS216_BINARY \
                    or binaryType == win32file.SCS_POSIX_BINARY:
                return BinaryTypes.Bit16
            else:
                return BinaryTypes.Bit64

    # TODO: Add for other OS
    return BinaryTypes.Bit32



import importlib
import sys
import os.path as path
from typing import Optional, Dict, Union

import yaml

import autoinsight.data
from autoinsight.common.models.Configuration import Configuration
from .ServiceBase import ServiceBase
from ..common.CustomTyping import Serializable


class ConfigurationServiceBase(ServiceBase):
    _outputDirName = "output"

    def __init__(self, *args, **kwargs):
        self._configFilename: str = f"{autoinsight.__name__}.yml"
        self._scriptConfig: Optional[Dict] = None
        self._userConfig: Optional[Dict] = None
        builtinConfigContent: str = importlib.resources.read_text(autoinsight.data, self.configFileName)
        self._builtInConfig: Dict = self.load(builtinConfigContent)

        self._scriptConfigPath: Optional[str] = None
        self._config: Optional[Configuration] = None
        self._scriptRuntimeConfig: Optional[Dict[str, Serializable]] = None
        self._scriptName: str = ""

    @property
    def configFileName(self) -> str:
        return self._configFilename

    @property
    def scriptName(self) -> str:
        return self._scriptName

    @property
    def config(self) -> Configuration:
        if not self._config:
            self._config = Configuration(**self.loadAllConfigurations(self._scriptRuntimeConfig))

        return self._config

    @property
    def outputRoot(self) -> str:
        """
        If there is no output location specific by the configuration and the script then
        the default output location will be point to the entry point of the running process
        """
        if self.config.script.output_root:
            return self.config.script.output_root
        else:
            return path.join(path.dirname(sys.argv[0]), self._outputDirName)

    @property
    def builtinConfig(self) -> Dict:
        return self._builtInConfig

    @classmethod
    def load(cls, configData: str) -> Optional[Dict]:
        """
        Load the yaml configuration file from the configData. It's either a file pat or the yaml content
        """
        try:
            if path.isfile(configData):
                return yaml.load(open(configData), yaml.Loader)
            elif configData:
                return yaml.load(configData, yaml.Loader)
        except:
            return

    def loadAllConfigurations(self, *configInfo: Union[str, Dict]) -> Dict:
        """
        Reload the configuration file from default and custom location
        """
        configPaths, configs = [], []
        for c in configInfo:
            if isinstance(c, str):
                configPaths.append(c)
            else:
                configs.append(c)

        loadedConfigs = (self.load(c) for c in configPaths)
        config: Dict = self.merge(self.builtinConfig,
                                  self._userConfig,
                                  self._scriptConfig,
                                  *loadedConfigs,
                                  *configs)
        return config

    @classmethod
    def merge(cls, *configs: Optional[Dict]) -> Dict:
        """
        Merge multiple levels of configurations into a single one. The later will
        overwrite the previous settings.
        """
        merged = {}
        for config in configs:
            if config and isinstance(config, Dict):
                merged.update(config)

        return merged

    def updateConfig(self, script):
        """
        Update the configuration base on the provided script information
        """
        self._scriptConfigPath = path.join(script.location, f"{autoinsight.__name__}.yml")
        self._scriptConfig = self.load(self._scriptConfigPath)
        self._scriptRuntimeConfig: Optional[Dict[str, Serializable]] = script.runtimeConfig
        self._scriptName = script.name
        config = Configuration(**self.loadAllConfigurations(script.runtimeConfig))

        if not config.common.log_format:
            config.common.log_format = "%(asctime)s - %(levelname)s: %(message)s"

        if not config.script.output_root:
            config.script.output_root = path.join(script.location, self._outputDirName)

        self._config = config



import collections
import sys
from typing import Tuple

from pywinauto import findbestmatch
from pywinauto.application import WindowSpecification

from autoinsight.common.CustomTyping import ElementsInfo


# TODO set the default value base on configuration
def _get_elements_info(self, depth=50, max_width=30) -> ElementsInfo:
    """
    Get all the UI elements

    The elements in the control and for its descendants to
    a depth of **depth** (the whole subtree if **None**).

    :param self: The instance of the WindowSpecification.

    :param depth: Max depth level of an element tree to dump (None: unlimited).

    :param max_width: Max number of children of each element to dump (None: unlimited).

    .. note:: The identifiers dumped by this method have been made
           unique. So if you have 2 edit boxes, they won't both have "Edit"
           listed in their identifiers. In fact the first one can be
           referred to as "Edit", "Edit0", "Edit1" and the 2nd should be
           referred to as "Edit2".
    """
    if depth is None:
        depth = sys.maxsize
    if max_width is None:
        max_width = sys.maxsize
    # Wrap this control
    this_ctrl = self._WindowSpecification__resolve_control(self.criteria)[-1]

    ElementTreeNode = collections.namedtuple('ElementTreeNode', ['elem', 'id', 'children'])

    def create_element_tree(element_list) -> Tuple[ElementTreeNode, bool, bool]:
        """Build elements tree and create list with pre-order tree traversal"""
        depth_limit_reached = False
        width_limit_reached = False
        current_id = 0
        elem_stack = collections.deque([(this_ctrl, None, 0)])
        root_node = ElementTreeNode(elem=this_ctrl, id=current_id, children=[])
        while elem_stack:
            current_elem, current_elem_parent_children, current_node_depth = elem_stack.pop()
            if current_elem is None:
                elem_node = ElementTreeNode(None, current_id, [])
                current_elem_parent_children.append(elem_node)
            else:
                if current_node_depth <= depth:
                    if current_elem_parent_children is not None:
                        current_id += 1
                        elem_node = ElementTreeNode(current_elem, current_id, [])
                        current_elem_parent_children.append(elem_node)
                        element_list.append(current_elem)
                    else:
                        elem_node = root_node
                    child_elements = current_elem.children()
                    if len(child_elements) > max_width and current_node_depth < depth:
                        elem_stack.append((None, elem_node.children, current_node_depth + 1))
                        width_limit_reached = True
                    for i in range(min(len(child_elements) - 1, max_width - 1), -1, -1):
                        elem_stack.append((child_elements[i], elem_node.children, current_node_depth + 1))
                else:
                    depth_limit_reached = True
        return root_node, depth_limit_reached, width_limit_reached

    # Create a list of this control, all its descendants
    all_ctrls = [this_ctrl]

    # Build element tree
    elements_tree, depth_limit_reached, width_limit_reached = create_element_tree(all_ctrls)
    txt_ctrls = None
    all_ctrl_index_names_map = None

    if self.allow_magic_lookup:
        # Create a list of all visible text controls
        txt_ctrls = [ctrl for ctrl in all_ctrls if ctrl.can_be_label and ctrl.is_visible() and ctrl.window_text()]

        # Build a dictionary of disambiguated list of control names
        name_ctrl_id_map = findbestmatch.UniqueDict()
        for index, ctrl in enumerate(all_ctrls):
            ctrl_names = findbestmatch.get_control_names(ctrl, all_ctrls, txt_ctrls)
            for name in ctrl_names:
                name_ctrl_id_map[name] = index

        # Swap it around so that we are mapped off the control indices
        all_ctrl_index_names_map = {}
        for name, index in name_ctrl_id_map.items():
            all_ctrl_index_names_map.setdefault(index, []).append(name)

    return ElementsInfo(ctrlTreeRoot=elements_tree,
                        textCtrls=txt_ctrls,
                        allCtrlIndexNameMaps=all_ctrl_index_names_map,
                        allCtrl=all_ctrls)


WindowSpecification.get_elements_info = _get_elements_info



from os import environ, path

from .ConfigurationServiceBase import ConfigurationServiceBase


class WindowConfigurationService(ConfigurationServiceBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._userConfigPath: str = path.join(environ["userprofile"], self.configFileName)
        self._userConfig = self.load(self._userConfigPath)



from dataclasses import dataclass

from .Rectangle import Rectangle
from .ModelBase import ModelBase


@dataclass
class IdentResult(ModelBase):
    rect: Rectangle
    confidence: float
    classnum: int

    def __str__(self):
        return f"rect: {self._rect}, confidence: {self._confidence}, class_num: {self._class_num}"



import platform
from typing import Type, Optional, Dict, Any

from autoinsight.common.EnumTypes import OSTypes
from .ConfigurationServiceBase import ConfigurationServiceBase
from .ContextManagementService import ContextManagementService
from .KnowledgeServiceBase import KnowledgeServiceBase
from .OCRServiceBase import OCRServiceBase
from .ServiceBase import ServiceBase
from .TesseractOCRService import TesseractOCRService
from .AIServiceBase import AIServiceBase
from .YoloAIService import YoloAIService
from .WindowConfigurationService import WindowConfigurationService


class IoCService(ServiceBase):
    OSType = OSTypes.fromStr(platform.system())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._registeredServices: Dict[OSTypes, Dict[Any, Any]] = {
            OSTypes.Any: {
                OCRServiceBase: TesseractOCRService(),
                AIServiceBase: YoloAIService(),
                ContextManagementService: ContextManagementService()
            },
            OSTypes.Windows: {},
            OSTypes.Linux: {},
            OSTypes.MacOS: {}
        }
        self._registerWindowServices()

    def getService(self, serviceType: Type[ServiceBase]) -> Optional[Type[ServiceBase]]:
        try:
            rss = self._registeredServices
            if serviceType in rss[OSTypes.Any]:
                return rss[OSTypes.Any][serviceType]

            return rss[self.OSType][serviceType]
        except KeyError:
            return

    def _registerWindowServices(self):
        if self.OSType == OSTypes.Windows:
            from autoinsight.services.WindowKnowledgeService import WindowKnowledgeService
            from wmi import WMI

            serviceMap = self._registeredServices[OSTypes.Windows]
            serviceMap.setdefault(KnowledgeServiceBase, WindowKnowledgeService())
            serviceMap.setdefault(ConfigurationServiceBase, WindowConfigurationService())
            serviceMap.setdefault(WMI, WMI())



from abc import abstractmethod
from typing import Optional, Sequence

from .ServiceBase import ServiceBase
from autoinsight.common.models.Rectangle import Rectangle
from autoinsight.common.models.Size import Size


class ImageProcessingServiceBase(ServiceBase):
    @abstractmethod
    def getImageResolution(self, image) -> Optional[Size]:
        pass

    @abstractmethod
    def isImageCorrupted(self, image) -> bool:
        pass

    @abstractmethod
    def compareBrightness(self, imageA, imageB) -> int:
        pass

    @abstractmethod
    def getBSODInfo(self, image) -> str:
        pass

    @abstractmethod
    def isVideoStutter(self, video) -> bool:
        pass

    @abstractmethod
    def isVideoFlicker(self, video) -> bool:
        pass

    @abstractmethod
    def isContainObject(self, image, target) -> bool:
        pass

    @abstractmethod
    def compareSharpness(self, imageA, imageB) -> int:
        pass

    @abstractmethod
    def compareNoise(self, imageA, imageB) -> int:
        pass

    @abstractmethod
    def isPrivacyLEDOn(self, imageA, imageB) -> bool:
        pass

    @abstractmethod
    def getImageColorTemperature(self, image) -> int:
        pass

    @abstractmethod
    def getVideoColorTemperature(self, video, checkpoint) -> int:
        pass

    @abstractmethod
    def getImageHSVMean(self, image) -> Sequence[int]:
        pass

    @abstractmethod
    def getVideoHSVMean(self, video, checkpoint) -> int:
        pass

    @abstractmethod
    def getImageRGBMean(self, image) -> Sequence[int]:
        pass

    @abstractmethod
    def getVideoRGBMean(self, video, checkpoint) -> Sequence[int]:
        pass

    @abstractmethod
    def compareFocusResult(self, imageA, imageB, rectangle) -> int:
        pass

    @abstractmethod
    def getBrightestRegion(self, image) -> Rectangle:
        pass

    @abstractmethod
    def getDarkestRegion(self, image) -> Rectangle:
        pass

    @abstractmethod
    def getImageISO(self, image) -> int:
        pass

    @abstractmethod
    def getVideoISO(self, video, checkpoint) -> int:
        pass

    @abstractmethod
    def compareImageISO(self, imageA, imageB) -> int:
        pass

    @abstractmethod
    def compareVideoISO(self, videoA, videoB, checkpoint) -> int:
        pass

    @abstractmethod
    def compareImageContrast(self, imageA, imageB) -> int:
        pass

    @abstractmethod
    def compareVideoContrast(self, videoA, videoB, checkpoint) -> int:
        pass

    @abstractmethod
    def compareImageGamma(self, imageA, imageB) -> int:
        pass

    @abstractmethod
    def compareVideoGamma(self, videoA, videoB, checkpoint) -> int:
        pass



from __future__ import annotations
from itertools import chain

from typing import Any, Optional, Dict, Sequence

from .Step import Step
from autoinsight.decorator.Log import log, Log


class Record(Step):
    _records: Dict[str, list[Record]] = {}

    def __init__(self, value: Any, recordKey: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._value = value
        self._recordKey = str
        self._addRecord(recordKey, self)

    @log
    def execute(self, isPaused: bool = False):
        Log.logger.info("Execute %s", self)
        self._isExecuted = True

    @property
    def value(self) -> Any:
        return self._value

    @classmethod
    @log
    def records(cls, recordKey: Optional[str] = None) -> Sequence[Record]:
        if recordKey and recordKey in cls._records:
            return cls._records[recordKey]
        else:
            return tuple(chain(*cls._records.values()))

    @classmethod
    @log
    def _addRecord(cls, key: str, record: Record):
        if key in cls._records:
            cls._records[key].append(record)
        else:
            cls._records.setdefault(key, [record])



from .GUIApplicationBase import GUIApplicationBase


class BrowserBase(GUIApplicationBase):
    @property
    def tabs(self):
        pass

    def navigate(self, url: str):
        pass

    def focus(self):
        pass

    def close(self):
        pass

    def closeTab(self):
        pass



from typing import Optional

from .ModelBase import ModelBase
from autoinsight.common.EnumTypes import EventTypes
from autoinsight.common.CustomTyping import Serializable


class Event(ModelBase):
    type: EventTypes
    args: Serializable
    error: Optional[BaseException]



from .TargetBase import TargetBase


class Text(TargetBase):
    def __init__(self, value="", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    def __repr__(self) -> str:
        pass

    def __str__(self) -> str:
        pass



from abc import abstractmethod
from typing import Iterable, Optional

from .ProcessBase import ProcessBase
from autoinsight.common.models.Point import Point
from autoinsight.common.models.Rectangle import Rectangle
from autoinsight.common.models.Size import Size
from autoinsight.ident.target.TargetBase import TargetBase
from autoinsight.common.CustomTyping import AutomationInstance


class FormBase(ProcessBase):
    def __init__(self,
                 automationInstance: Optional[AutomationInstance] = None,
                 parent: Optional[ProcessBase] = None,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self._x = 0
        self._y = 0
        self._width = 0
        self._height = 0
        self._parent: Optional[ProcessBase] = parent
        self._automationInstance = automationInstance

    @property
    def automationInstance(self) -> Optional[AutomationInstance]:
        return self._automationInstance

    @property
    def title(self) -> str:
        if self.automationInstance:
            self.automationInstance.tite()

    @property
    def position(self) -> Point:
        return Point(x=self._x, y=self._y)

    @property
    def parent(self) -> Optional[ProcessBase]:
        return self._parent or super().parent

    @property
    def targets(self) -> Iterable[TargetBase]:
        pass

    @property
    def size(self) -> Size:
        return Size(width=self._width, height=self._height)

    @property
    def rectangle(self) -> Rectangle:
        return Rectangle(left=self._x, top=self._y, width=self._width, height=self._height)

    @abstractmethod
    def focus(self):
        pass

    @abstractmethod
    def maximize(self):
        pass

    @abstractmethod
    def minimize(self):
        pass

    @abstractmethod
    def drag(self):
        pass

    @abstractmethod
    def drop(self):
        pass

    @abstractmethod
    def dragTo(self, x: int, y: int):
        pass

    @abstractmethod
    def snapshot(self):
        pass

    @abstractmethod
    def scroll(self):
        pass

    @abstractmethod
    def isDraggable(self):
        pass



from __future__ import annotations

import time
from typing import Optional, Iterable, Union

from PIL import Image

from autoinsight.common.CustomTyping import AutomationInstance
from autoinsight.common.models.Point import Point
from autoinsight.common.models.Rectangle import Rectangle
from autoinsight.common.models.Size import Size
from autoinsight.common.EnumTypes import KnownContextQueries
from autoinsight.ident.IdentObjectBase import IdentObjectBase
from autoinsight.services.ContextManagementService import ContextManagementService
from autoinsight.services.IoCService import IoCService
from autoinsight.decorator.Log import log


class TargetBase(IdentObjectBase):
    def __init__(self,
                 query: str,
                 ioc: IoCService = IoCService(),
                 automationInstance: Optional[AutomationInstance] = None,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self._cms: ContextManagementService = ioc.getService(ContextManagementService)
        self._x = 0
        self._y = 0
        self._width = 0
        self._height = 0
        self._automationInstance: AutomationInstance = automationInstance
        self._query: Union[str, KnownContextQueries] = query
        self._alias = []

    @property
    def description(self) -> str:
        return self._query

    @property
    def aliases(self) -> Iterable[str]:
        """
        Return the alias type names of the current target for example:
        edit, input
        """
        return self._alias

    @property
    def automationInstance(self):
        if not self._automationInstance and self._query:
            self._automationInstance = self._cms.currentContext.find(self._query, self)
        return self._automationInstance

    @property
    def size(self) -> Size:
        return Size(width=self._width, height=self._height)

    @property
    def rectangle(self) -> Rectangle:
        """
        Relative left, top location base on the current context
        """
        if self.isExist():
            # 'bottom', 'height', 'left', 'mid_point', 'right', 'top', 'width'
            r = self.automationInstance.rectangle()
            self._x = r.left
            self._y = r.top
            self._width = r.width
            self._height = r.height
            return Rectangle(left=self._x, top=self._y, width=self._width, height=self._height)
        else:
            return Rectangle(left=0, top=0, width=0, height=0)

    @property
    def center(self) -> Point:
        return self.rectangle.center

    @property
    def screenCenter(self) -> Point:
        """
        Return the center point of the target base on the screen coordinate
        """
        if self.isExist():
            return self.automationInstance.client_to_screen(self.center)
        else:
            return Point(x=0, y=0)

    @property
    def parent(self) -> Optional[TargetBase]:
        return self._cms.currentContext

    @log
    def click(self) -> bool:
        if self.isExist():
            try:
                self.automationInstance.wait_for_idle()
                self.automationInstance.set_focus()
                self.automationInstance.click_input()
                return True
            except:
                return False

    @log
    def rightClick(self):
        if self.isExist():
            try:
                self.automationInstance.right_click_input()
                return True
            except:
                return False

    @log
    def doubleRightClick(self):
        if self.isExist():
            try:
                self.automationInstance.double_click_input()
                return True
            except:
                return False

    @log
    def doubleClick(self):
        if self.isExist():
            try:
                self.automationInstance.double_click_input()
                return True
            except:
                return False

    @log
    def drag(self):
        pass

    @log
    def drop(self):
        pass

    @log
    def isVisible(self) -> bool:
        if self.isExist():
            try:
                return self.automationInstance.is_visible()
            except:
                return False

    @log
    def isEnable(self) -> bool:
        if self.isExist():
            try:
                return self.automationInstance.is_enabled()
            except:
                return False

    @log
    def highlight(self):
        if self.isExist():
            try:
                self.parent.focus()
                # TODO make the wait time configurable
                for i in range(50):
                    self.automationInstance.draw_outline()
                    return True
            except:
                return False

    @log
    def mouseHover(self):
        if self.isExist():
            self.automationInstance.client_to_screen(self.center)

    @log
    def scrollIntoView(self, timeout: int = 5) -> bool:
        # TODO extend to support scroll down, up, left and right etc.
        parent = self._getScrollableParent()
        startTime = time.time()
        isTimeout = False
        while not self.isVisible() and parent:
            parent().scroll("down", "line")

            if time.time() - startTime > timeout:
                isTimeout = True
                break

        return not isTimeout

    @log
    def reFind(self):
        self._automationInstance = None

    @log
    def isScrollable(self):
        try:
            if not self.isExist():
                return False
            else:
                return not not self.automationInstance.iface_scroll
        except:
            return False

    @log
    def _getScrollableParent(self) -> Optional[TargetBase]:
        parent = self.parent

        while parent:
            if parent.isScrollable():
                return parent

            parent = parent.parent

    @log
    def snapshot(self) -> Image:
        if self.isExist():
            image: Image = self.automationInstance.capture_as_image()
            return image

    @log
    def isExist(self) -> bool:
        """
        Check the target can be found on the current context or not
        """
        return bool(self.automationInstance)

    @log
    def waitFor(self, timeout: int = 10, interval: int = 1) -> bool:
        """
        Wait for the target exist till timeout
        """
        start = time.time()
        while not self.isExist():
            time.sleep(interval)
            if time.time() - start > timeout:
                break

        return self.isExist()



from .TargetBase import TargetBase
from autoinsight.decorator.Log import log
from autoinsight.decorator.Step import step
from autoinsight.ident.context.WindowStartMenu import WindowStartMenu
from autoinsight.common.EnumTypes import KnownContextQueries
from autoinsight.ident.automation.WindowStartMenuAuto import WindowStartMenuAuto


class StartMenu(TargetBase):
    def __init__(self, *args, **kwargs):
        super().__init__(KnownContextQueries.WindowStartMenu.name, *args, **kwargs)
        self._alias.extend(["start", "StartMenu", "StartButton"])
        self._isToggled = False
        self._query = KnownContextQueries.WindowStartMenu

    @property
    def description(self) -> str:
        return KnownContextQueries.WindowStartMenu.name

    def __repr__(self):
        return

    def __str__(self):
        pass

    @property
    def automationInstance(self) -> WindowStartMenuAuto:
        if not self._automationInstance:
            self._cms.currentContext = WindowStartMenu()
            self._automationInstance = self._cms.currentContext.find(self._query, self)
        return self._automationInstance

    def _toggle(self) -> bool:
        if self.automationInstance:
            try:
                self.automationInstance.toggle()
                self._isToggled = not self._isToggled
                return True
            except:
                return False

    @step
    def click(self) -> bool:
        return self._toggle()

    @step
    def toggle(self) -> bool:
        return self._toggle()

    @log
    def isToggleButton(self) -> bool:
        return True

    @log
    def isExist(self) -> bool:
        return True



from abc import abstractmethod
from typing import Any

from autoinsight.common.ObjectBase import ObjectBase
from autoinsight.ident.IdentObjectBase import IdentObjectBase
from autoinsight.services.IoCService import IoCService
from autoinsight.services.ContextManagementService import ContextManagementService
from autoinsight.common.Utils import GUID


class StepBase(ObjectBase):
    def __init__(self,
                 name: str,
                 action: str,
                 target: IdentObjectBase,
                 context: IdentObjectBase,
                 description: str = "",
                 *args,
                 **kwargs):
        self._name = name
        self._action = action
        self._isExecuted = False
        self._result = None
        self._target: IdentObjectBase = target
        self._context: IdentObjectBase = context
        self._description = description
        self._os = None
        self._id = GUID()

    @property
    def isExecuted(self):
        return self._isExecuted

    @property
    def os(self):
        if not self._os:
            cms: ContextManagementService = IoCService().getService(ContextManagementService)
            self._os = cms.os
        return self._os

    @property
    def result(self) -> Any:
        return self._result

    @property
    def name(self) -> str:
        return self._name

    @property
    def action(self) -> str:
        return self._action

    def reset(self):
        """
        Reset the step to init state
        """
        self._isExecuted = False

    @abstractmethod
    def execute(self, isPaused: bool = False):
        pass

    @abstractmethod
    def __repr__(self) -> str:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass



from time import sleep

from autoinsight.script.Step import Step
from autoinsight.ident.IdentObjectBase import IdentObjectBase


class Wait(Step):
    def __init__(self, second: int, *args, **kwargs):
        self._second = second
        super().__init__('Wait', f"second {second}", IdentObjectBase(), IdentObjectBase())

    def execute(self, *args, **kwargs):
        sleep(self._second)



from .ControlBase import ControlBase
from autoinsight.decorator.Log import log
from autoinsight.decorator.Step import step


class Hyperlink(ControlBase):
    def __init__(self, query: str, *args, **kwargs):
        super().__init__(query, *args, **kwargs)

    def __repr__(self):
        return

    def __str__(self):
        pass

    @log
    @step
    def click(self) -> bool:
        if super().click():
            return True
        else:
            return False

    @property
    def text(self) -> str:
        return self._text



import collections
from typing import Union, NewType, Optional, Dict, Callable
from pathlib import Path

import numpy as np
from PIL.Image import Image

from pywinauto.application import WindowSpecification
from pywinauto.uia_element_info import UIAElementInfo
# TODO the imports should be replaced with the wrapper in autoinsight.automation.*
from pywinauto.win32_element_info import HwndElementInfo

# TODO in the future we need to separate the implementation for technical instance and visual instance
# Implement __getitem__ etc to simplified the try catch calling
AutomationInstance = NewType("AutomationInstance", Optional[Union[UIAElementInfo,
                                                                  HwndElementInfo,
                                                                  WindowSpecification]])

ElementTreeNode = collections.namedtuple("ElementTreeNode", ["elem", "id", "children"])
ElementsInfo = collections.namedtuple("ElementsInfo", ["ctrlTreeRoot",
                                                       "textCtrls",
                                                       "allCtrlIndexNameMaps",
                                                       "allCtrl"])
FoundCtrlItem = collections.namedtuple("FoundCtrlItem", ["ctrl", "score", "secondScore", "index"])

Serializable = NewType("Serializable", Union[bool,
                                             int,
                                             float,
                                             complex,
                                             str,
                                             bytes,
                                             bytearray,
                                             tuple,
                                             list,
                                             set,
                                             frozenset,
                                             Dict,
                                             Callable])

IMG = NewType("IMG", Union[str, Image, np.array])
PathStr = NewType("PathStr", Union[Path, str])




import os


import autoinsight.extend.ExtendEnum
import autoinsight.extend.ExtendPywinautoApplicationWindowSpecification
from autoinsight.ident.target.Button import Button
from autoinsight.ident.target.ComboBox import ComboBox
from autoinsight.ident.context.WindowOS import WindowOS
from autoinsight.ident.target.StartMenu import StartMenu
from autoinsight.ident.target.SearchBox import SearchBox
from autoinsight.script.Wait import Wait
from autoinsight.ident.context.WindowGUIApplication import WindowGUIApplication

__version__ = "0.1.0"
packageRoot = os.path.dirname(__file__)
packageDataRoot = os.path.join(packageRoot, "data")



import logging
from abc import abstractmethod
from logging import FileHandler, Handler, Logger
from typing import Iterable


class LoggerBase:
    levelMaps = {
        "critical": logging.CRITICAL,
        "error": logging.ERROR,
        "warning": logging.WARNING,
        "info": logging.INFO,
        "debug": logging.DEBUG
    }

    def __init__(self, name: str, level: str, formatTemplate: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        logLevel = self._toLogLevel(level)
        logging.basicConfig(format=formatTemplate,
                            level=logLevel)
        self._logger: Logger = logging.getLogger(name)

        for h in self._handlers():
            h.setLevel(logLevel)
            h.setFormatter(logging.Formatter(formatTemplate))
            self._logger.addHandler(h)

    @classmethod
    def _toLogLevel(cls, level: str) -> int:
        return cls.levelMaps.get(level.lower(), logging.WARNING)

    @abstractmethod
    def _handlers(self) -> Iterable[Handler]:
        pass

    def debug(self, *args, **kwargs):
        self._logger.debug(*args, **kwargs)

    def info(self, *args, **kwargs):
        self._logger.info(*args, **kwargs)

    def warning(self, *args, **kwargs):
        self._logger.warning(*args, **kwargs)

    def error(self, *args, **kwargs):
        self._logger.error(*args, **kwargs)

    def critical(self, *args, **kwargs):
        self._logger.critical(*args, **kwargs)

    @staticmethod
    def close():
        logging.shutdown()


class ConsoleLogger(LoggerBase):
    def _handlers(self) -> Iterable[Handler]:
        return ()


class ConsoleAndFileLogger(LoggerBase):
    def __init__(self, logPath: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._logPath = logPath

    def _handlers(self) -> Iterable[Handler]:
        return (
            FileHandler(self._logPath, mode="w+"),
        )



# typing: ignore[misc]
from logging import Formatter
from typing import Literal, Optional

from pydantic import BaseModel

from .ModelBase import ModelBase


class ServiceConfig(BaseModel, ModelBase):
    pass


class IdentConfig(BaseModel, ModelBase):
    search_depth: int = 50
    search_width: int = 30
    wait_seconds_for_object: int = 30
    wait_seconds_for_window: int = 30
    wait_retry_interval_for_object: int = 3
    wait_retry_interval_for_window: int = 5
    highlight_wait_seconds: int = 10


class ScriptConfig(BaseModel, ModelBase):
    output_root: Optional[str] = None
    is_snapshot_before_action: bool = False
    action_delay_seconds: int = 0


class CommonConfig(BaseModel, ModelBase):
    log_format: Optional[str] = None
    log_level: Optional[Literal["critical", "error", "warning", "info", "debug"]] = "info"

    def __post_init__(self):
        if self.log_format:
            try:
                self._log_format: Formatter = Formatter(self.log_format)
            except ValueError:
                self.log_format = None


class Configuration(BaseModel, ModelBase):
    common: CommonConfig
    script: ScriptConfig
    ident: IdentConfig
    service: ServiceConfig



import os
import platform
import signal
from collections.abc import Sequence
from subprocess import Popen
from typing import Iterable, Mapping, Optional

import pyautogui
from pywinauto import WindowSpecification
from wmi import WMI, _wmi_object

from .Command import Command
from .OSBase import OSBase
from .PowerShell import PowerShell
from .ProcessBase import ProcessBase
from .WindowGUIContextBase import WindowGUIContextBase
from .WindowGUIApplication import WindowGUIApplication
from autoinsight.common.EnumTypes import ButtonTypes
from autoinsight.common.models.Point import Point
from autoinsight.services.KnowledgeServiceBase import KnowledgeServiceBase
from autoinsight.services.WindowKnowledgeService import WindowKnowledgeService
from autoinsight.decorator.Log import log, Log
from autoinsight.decorator.Step import step
from autoinsight.common.CustomTyping import AutomationInstance


class WindowOS(OSBase, WindowGUIContextBase):
    def __init__(self,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self._wmi: _wmi_object = self._ioc.getService(WMI)
        self._ks: WindowKnowledgeService = self._ioc.getService(KnowledgeServiceBase)

    def __repr__(self):
        if not self._repr:
            self._repr = f"{self.__str__()} id:{self.id} {self.version}"
        return self._repr

    def __str__(self):
        if not self._str:
            self._str = "Window OS Context"
        return self._str

    @property
    def processes(self) -> Iterable[ProcessBase]:
        pass

    @property
    def desktop(self):
        pass

    @property
    def description(self) -> str:
        return self.version

    @property
    def ks(self) -> WindowKnowledgeService:
        return self._ks

    @property
    def wmi(self) -> _wmi_object:
        return self._wmi

    @property
    def userHome(self):
        return self.environ["USERPROFILE"]

    @property
    def environ(self) -> Mapping[str, str]:
        return self._environ

    @property
    def monitors(self):
        pass

    @property
    def systeminfo(self) -> str:
        pass

    @property
    def version(self) -> str:
        return f"{platform.system()} {platform.version()} {platform.release()} {platform.win32_edition()}"

    @property
    def drivers(self) -> Iterable[str]:
        pass

    @property
    def disks(self) -> Iterable[str]:
        pass

    @property
    def SBIOS(self) -> str:
        pass

    @property
    def VBIOS(self) -> str:
        pass

    @property
    def services(self):
        pass

    @log
    @step
    def shutdown(self, delaySeconds: int = 0):
        delaySeconds: int = int(delaySeconds)
        os.system("shutdown /s /t %s" % delaySeconds)

    @log
    @step
    def holdAndTypeKeys(self, keys: Iterable[str], holdKeys: Sequence[str]):
        countOfControlKeys: int = len(holdKeys)
        if countOfControlKeys == 0:
            self.typeKeys(keys)
        elif countOfControlKeys == 1:
            with pyautogui.hold(holdKeys[0]):
                self.typeKeys(keys)
        elif countOfControlKeys == 2:
            with pyautogui.hold(holdKeys[0]):
                with pyautogui.hold(holdKeys[1]):
                    self.typeKeys(keys)
        elif countOfControlKeys == 3:
            with pyautogui.hold(holdKeys[0]):
                with pyautogui.hold(holdKeys[1]):
                    with pyautogui.hold(holdKeys[2]):
                        self.typeKeys(keys)

    @log
    @step
    def changeDriverState(self, driverInfo: str, isEnable: bool):
        pass

    @log
    @step
    def changeServiceState(self, serviceInfo: str, isEnable: bool):
        pass

    @log
    @step
    def install(self, path: str):
        """Automatic install an application
        """
        pass

    @log
    @step
    def typeControlKeys(self, controlKeys: Iterable[str]):
        """Send simulated control keys

        Args:
            controlKeys (str): a string contain the control keys like {ENTER} etc.
        """
        pyautogui.write(controlKeys)

    @log
    @step
    def hotkeys(self, *keys):
        pyautogui.hotkey(*keys)

    @log
    @step
    def querySystemEvent(self, eventPattern: str, level):
        pass

    @log
    @step
    def restart(self, delaySeconds: int = 0):
        delaySeconds: int = int(delaySeconds)
        os.system("shutdown /s /r /t %s" % delaySeconds)

    @log
    @step
    def sleep(self):
        os.system("POWERCFG /HIBERNATE OFF")
        os.system("RUNDLL32.EXE powrprof.dll,SetSuspendState 0,1,0")

    @log
    @step
    def hibernate(self):
        os.system("POWERCFG /HIBERNATE ON")
        os.system("shutdown /s /hybrid")

    # TODO: Refactor the knowledge into the WindowGUIApplication initialization parameters
    @log
    @step
    def launchApp(self, cmdline: str, isAsAdmin: bool = False, *args, **kwargs) -> WindowGUIApplication:
        instance = self.ks.recognize(cmdline)
        Log.logger.info("Launch application %s as admin %s", cmdline, isAsAdmin)
        if instance:
            return WindowGUIApplication(automationInstance=instance)
        else:
            return WindowGUIApplication(cmdline=cmdline, *args, **kwargs)

    @log
    @step
    def launchShell(self, cmds: Iterable[str], isAsAdmin: bool = False, *args, **kwargs) -> Optional[Command]:
        Log.logger.info("Launch shell %s as admin %s", cmds, isAsAdmin)
        return Command(cmds, isAsAdmin, *args, **kwargs)

    @log
    @step
    def launchShellWaitWindow(self, cmds: Sequence[str], windowQuery: str, isAsAdmin: bool = False, *args, **kwargs):
        Log.logger.info("Launch shell and wait window %s as admin %s", cmds, isAsAdmin)
        return self.waitForWindow(windowQuery, Popen(cmds))

    @log
    @step
    def launchPowershell(self, *args, **kwargs) -> Optional[PowerShell]:
        return PowerShell(*args, **kwargs)

    @log
    @step
    def launchFromStartMenu(self, appName: str, appQuery: str, *args, **kwargs) -> Optional[WindowGUIApplication]:
        pass

    @log
    @step
    def kill(self, processId: int = 0, processName: str = ""):
        if processId:
            os.kill(processId, signal.SIGKILL)

    @log
    @step
    def terminate(self, processId: int = 0, processName: str = ""):
        if processId:
            os.kill(processId, signal.SIGTERM)

    @log
    @step
    def updateDriver(self):
        pass

    @log
    @step
    def typeKeys(self, visibleKeys: Iterable[str], intervalSec: int = 0.25):
        """
        Simulate typing the visible characters
        """
        pyautogui.write(visibleKeys, interval=intervalSec)

    @log
    @step
    def snapshot(self):
        pass

    @log
    @step
    def select(self, start: Point, end: Point, button: ButtonTypes = ButtonTypes.Left):
        pyautogui.mouseDown(x=start.x, y=start.y, button=button.str)
        pyautogui.mouseUp(x=end.x, y=end.y, button=button.str)

    @log
    @step
    def switchUser(self):
        pass

    @log
    @step
    def logout(self):
        pass

    @log
    @step
    def waitForWindow(self, query: str, processHandle: Optional[Popen] = None) -> WindowGUIApplication:
        window: AutomationInstance = WindowSpecification({"backend": "uia", "best_match": query})

        window.wait("exists visible ready",
                    timeout=self._cs.config.ident.wait_seconds_for_window,
                    retry_interval=self._cs.config.ident.wait_retry_interval_for_window)
        return WindowGUIApplication(automationInstance=window, processHandle=processHandle)



from abc import abstractmethod

from autoinsight.common.ObjectBase import ObjectBase


class MonitorBase(ObjectBase):
    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def stop(self):
        pass



import os
import platform
from functools import partial
from multiprocessing import Process
from typing import Dict

from pywinauto import Desktop, WindowSpecification, Application

from autoinsight.common.CustomTyping import AutomationInstance
from autoinsight.common.EnumTypes import OSTypes
from autoinsight.common.models.Knowledge import Knowledge
from .KnowledgeServiceBase import KnowledgeServiceBase


class WindowKnowledgeService(KnowledgeServiceBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.knowledge.append(Knowledge(alias=["windowscamera",
                                               "windowscamera.exe",
                                               "camera",
                                               "camera.exe",
                                               "microsoft.windows.camera"],
                                        launch=_launchCamera))

        self.knowledge.append(Knowledge(alias=["rdp", "remote desktop"],
                                        launch=_launchRDP))

        self.knowledge.append(Knowledge(alias=["control panel"],
                                        launch=_launchControlPanel))

        self.knowledge.append(Knowledge(alias=["settings"],
                                        launch=_launchSettings))

        self.knowledge.append(Knowledge(alias=["calc", "calc.exe", "calculator"],
                                        launch=_launchCalc))

        self.knowledge.append(Knowledge(alias=["media player", "wmplayer.exe", "mediaplayer"],
                                        launch=_launchMediaPlayer))

    def recognize(self, cmdline: str) -> AutomationInstance:
        knowledge = None
        if cmdline:
            cmdline = cmdline.lower()
            for k in self.knowledge:
                if cmdline in k.alias:
                    knowledge = k
                    break

        if knowledge:
            return knowledge.launch(knowledge)

    def __repr__(self):
        pass

    def __str__(self):
        pass

    def wait(self, timeoutSeconds: int = 0):
        pass


def _task(commandline: str):
    os.system(commandline)


def _worker(commandline: str, k: Knowledge):
    p = Process(target=partial(_task, commandline), args=k.arguments)
    p.start()


def _waitWindow(criteria: Dict[str, str]) -> AutomationInstance:
    window = WindowSpecification(criteria)
    # TODO add magic numbers into global configuration
    window.wait("exists visible ready", timeout=30, retry_interval=3)
    return window


def _launchCamera(k: Knowledge) -> AutomationInstance:
    _worker('cmd /C "start microsoft.windows.camera: "', k)
    return _waitWindow({"backend": "uia", "best_match": "Camera"})


def _launchControlPanel(k: Knowledge) -> AutomationInstance:
    _worker('control.exe', k)
    return _waitWindow({"backend": "uia", "best_match": r"Control Panel"})


def _launchSettings(k: Knowledge) -> AutomationInstance:
    pass

def _launchRDP(k: Knowledge) -> AutomationInstance:
    Application(backend="uia").start(r"%windir%\system32\mstsc.exe")

    return _waitWindow({"backend": "uia", "best_match": "Remote Desktop Connection"})

def _launchCalc(k: Knowledge) -> AutomationInstance:
    Application(backend="uia").start('calc.exe')

    return Desktop(backend="uia").Calculator


def _launchMediaPlayer(k: Knowledge) -> AutomationInstance:
    path = rf'"{os.environ["HOMEDRIVE"]}\Program Files\Windows Media Player\wmplayer.exe"'
    _worker(path, k)
    return _waitWindow({"backend": "uia", "best_match": "Windows Media Player"})


def _isWindow11() -> bool:
    if OSTypes.fromStr(platform.system()) == OSTypes.Windows:
        version = int(platform.version().split(".")[-1])
        return version >= 22000
    else:
        return False



from abc import abstractmethod
from typing import Iterable, Optional

from .FormBase import FormBase
from .ProcessBase import ProcessBase


class GUIApplicationBase(ProcessBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._forms: Iterable[FormBase] = []
        self._currentForm: Optional[FormBase] = None

    @property
    def forms(self) -> Optional[Iterable[FormBase]]:
        pass

    @property
    def currentForm(self) -> Optional[FormBase]:
        return self._currentForm

    @abstractmethod
    def focus(self) -> bool:
        pass



from abc import ABC
from typing import Optional

from .TargetBase import TargetBase
from autoinsight.common.EnumTypes import ScrollDirectionTypes, ScrollAmountTypes
from autoinsight.ident.IdentObjectBase import IdentObjectBase
from autoinsight.decorator.Log import log
from autoinsight.decorator.Step import step


class ControlBase(TargetBase, ABC):
    @property
    def next(self) -> Optional[IdentObjectBase]:
        pass

    @property
    def previous(self) -> Optional[IdentObjectBase]:
        pass

    @log
    @step
    def focus(self) -> bool:
        if self.automationInstance:
            try:
                self.automationInstance.set_focus()
                return True
            except:
                return False

    @log
    @step
    def scroll(self, direction: ScrollDirectionTypes = ScrollDirectionTypes.Down):
        if isinstance(direction, str):
            direction: ScrollDirectionTypes = ScrollDirectionTypes.fromStr(direction)

        if self.isScrollable():
            self.automationInstance.scroll(direction.Down.str, ScrollAmountTypes.Lines.str)

    @log
    @step
    def wait(self, seconds: int):
        if self.parent:
            self.parent.wait(seconds)




from typing import Dict, Optional
from uuid import UUID

from autoinsight.ident.IdentObjectBase import IdentObjectBase
from .ServiceBase import ServiceBase


class ContextManagementService(ServiceBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._registeredContexts: Dict[UUID, IdentObjectBase] = {}
        self._currentContext: Optional[IdentObjectBase] = None
        self._contextQueue: list[IdentObjectBase] = []
        self._os: IdentObjectBase = None
        self._script: IdentObjectBase = None

    @property
    def currentContext(self) -> Optional[IdentObjectBase]:
        if not self._currentContext and self._contextQueue:
            self._currentContext = self._contextQueue[-1]
        return self._currentContext

    @currentContext.setter
    def currentContext(self, context: IdentObjectBase):
        if not context:
            self._currentContext = None

        if self._currentContext != context and context.id in self._registeredContexts:
            self._currentContext = context
            self._updateContextQueue(context)

    @property
    def os(self) -> Optional[IdentObjectBase]:
        return self._os

    @property
    def script(self) -> Optional[IdentObjectBase]:
        return self._script

    def registerContext(self, context: IdentObjectBase):
        if context.id not in self._registeredContexts:
            self._registeredContexts[context.id] = context
            self._contextQueue.append(context)

            if not self.currentContext:
                self.currentContext = context

    def _updateContextQueue(self, context: IdentObjectBase):
        if context.id in self._registeredContexts and self._contextQueue and self._contextQueue[-1] != context:
            self._contextQueue.remove(context)
            self._contextQueue.append(context)

    def unregisterContext(self, context: IdentObjectBase):
        if not context:
            return

        if context.id in self._registeredContexts:
            del self._registeredContexts[context.id]
            self._contextQueue.remove(context)

            if self.currentContext == context:
                self.currentContext = None

    def reset(self):
        """
        Clean all the registered contexts
        """
        self._registeredContexts.clear()
        self._contextQueue.clear()
        self._currentContext = None



from abc import abstractmethod
from typing import Iterable

from .ServiceBase import ServiceBase
from autoinsight.common.models.IdentResult import IdentResult


class AIServiceBase(ServiceBase):
    @abstractmethod
    def predict(self, image) -> Iterable[IdentResult]:
        pass



from abc import abstractmethod
from functools import update_wrapper
from typing import Callable

from autoinsight.common.ObjectBase import ObjectBase


class DecoratorBase(ObjectBase):
    def __init__(self, func: Callable, *args, **kwargs):
        self._func: Callable = func
        update_wrapper(self, func)

    @abstractmethod
    def __call__(self, *args, **kwargs):
        pass



from abc import ABC
from typing import Optional, Iterable, Any
from time import sleep

from autoinsight.common.CustomTyping import AutomationInstance, ElementsInfo
from autoinsight.common.models.Rectangle import Rectangle
from autoinsight.common.EnumTypes import GUIContainerTypes
from autoinsight.decorator.Log import log, Log
from autoinsight.ident.IdentObjectBase import IdentObjectBase
from autoinsight.services.ConfigurationServiceBase import ConfigurationServiceBase
from autoinsight.services.ContextManagementService import ContextManagementService
from autoinsight.services.IoCService import IoCService


class GUIContextBase(IdentObjectBase, ABC):
    def __init__(self,
                 ioc: IoCService = IoCService(),
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self._ioc: IoCService = ioc
        self._cms: ContextManagementService = ioc.getService(ContextManagementService)
        self._cms.registerContext(self)
        self._cs: ConfigurationServiceBase = ioc.getService(ConfigurationServiceBase)
        self._searchDepth: int = 500
        self._searchWidth: int = 300

    def __enter__(self):
        self.setCurrent()

    def __exit__(self, *args):
        self.tearDown()

    def __collectChildCtrls(self, ctrl) -> Iterable[Any]:
        pass

    @property
    def containers(self):
        return (c.name for c in GUIContainerTypes)

    @property
    def rectangle(self) -> Rectangle:
        return self._rectangle

    @property
    def automationInstance(self):
        return self._automationInstance

    def _updateRectangle(self):
        if self.automationInstance:
            rect = self.automationInstance.element_info.rectangle
            self._rectangle = Rectangle(rect.left, rect.top, rect.width(), rect.height())

    @log
    def setCurrent(self):
        self._cms.currentContext = self
        self._updateRectangle()

    @log
    def tearDown(self):
        self._cms.unregisterContext(self)

    @log
    def find(self, query: str, target: IdentObjectBase = None, *args, **kwargs) -> Optional[AutomationInstance]:
        pass

    @log
    def addDialogHandler(self, dialogQuery: str, ActionQuery: str, key: Optional[str] = None):
        """
        Create a background thread to handle random popup dialog
        """
        pass

    @log
    def addOneTimeDialogHandler(self, dialogQuery: str, ActionQuery: str, key: Optional[str] = None):
        """
        """

    @log
    def wait(self, seconds: int):
        sleep(seconds)



from typing import Tuple, Optional, Literal, Iterable

import pytesseract
from PIL import Image
from numpy import array

from autoinsight.common.models.Box import Box
from autoinsight.common.models.BoundingBox import BoundingBox
from .OCRServiceBase import OCRServiceBase


class TesseractOCRService(OCRServiceBase):

    def getTextFromImageFile(self, imagePath) -> str:
        return pytesseract.image_to_string(Image.open(imagePath))

    def getTextFromImageBinary(self, imageBinary: bytes, mode: str, size: Tuple[int, int]) -> str:
        mode: Literal["1", "CMYK", "F", "HSV", "I", "L", "LAB", "P", "RGB", "RGBA", "RGBX", "YCbCr"] = mode
        return pytesseract.image_to_string(Image.frombuffer(mode, size, imageBinary))

    def getTextFromImage(self, image: Image) -> str:
        return pytesseract.image_to_string(image)

    def getTextFromImageUrl(self, imageUrl: str, payload: str, auth: Optional[Tuple[str, str]] = None) -> str:
        image = self._getImageFromURL(auth, imageUrl, payload)
        return pytesseract.image_to_string(image)

    def getTextFromImageArray(self, imageArray: array) -> str:
        return pytesseract.image_to_string(array)

    def getBoxFromImageFile(self, imagePath) -> Optional[list[BoundingBox]]:
        return pytesseract.image_to_boxes(Image.open(imagePath))

    def getBoxFromImageBinary(self,
                              imageBinary: bytes,
                              mode: str,
                              size: Tuple[int, int]) -> Optional[list[BoundingBox]]:
        mode: Literal["1", "CMYK", "F", "HSV", "I", "L", "LAB", "P", "RGB", "RGBA", "RGBX", "YCbCr"] = mode
        return pytesseract.image_to_boxes(Image.frombuffer(mode, size, imageBinary))

    def getBoxFromImage(self, image: Image) -> Optional[list[BoundingBox]]:
        return self._parseBox(pytesseract.image_to_boxes(image))

    def getBoxFromImageUrl(self, imageUrl: str,
                           payload: str = None,
                           auth: Tuple[str, str] = None) -> Optional[list[BoundingBox]]:
        image = self._getImageFromURL(auth, imageUrl, payload)
        return pytesseract.image_to_boxes(image)

    def getBoxFromImageArray(self, imageArray: array) -> Optional[list[BoundingBox]]:
        return pytesseract.image_to_boxes(array)

    @classmethod
    def _parseBox(cls, boxesDescription: str) -> Optional[list[BoundingBox]]:
        if boxesDescription:
            boxes: list = []
            for line in boxesDescription.split("\n"):
                if line:
                    items: Iterable[str] = line.split(" ")
                    char = items[0]
                    left, top, width, height = [int(item) for item in items[1:5]]
                    boxes.append(BoundingBox(char,
                                             Box(left, top, width, height)))
            return boxes



from typing import Iterable, Optional

from PIL.Image import Image
from pywinauto import Application, Desktop

from .FormBase import FormBase
from .GUIApplicationBase import GUIApplicationBase
from .WindowGUIContextBase import WindowGUIContextBase
from .ProcessBase import ProcessBase
from .WindowForm import WindowForm
from autoinsight.decorator.Log import log
from autoinsight.decorator.Step import step
from autoinsight.common.CustomTyping import AutomationInstance


class WindowGUIApplication(GUIApplicationBase, WindowGUIContextBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._description = ""

    @classmethod
    def new(cls, *args, **kwargs) -> ProcessBase:
        return WindowGUIApplication(*args, **kwargs)

    @property
    def description(self) -> str:
        if not self._description:
            if self.processName:
                self._description += f"name:{self.processName}"
            elif self.processId:
                self._description += f"pid:{self.processId}"
            elif self._title:
                self._description += f"title:{self._title}"
            elif self._cmdline:
                self._description += f"cmdline:{self._cmdline}"

        return self._description

    @property
    def forms(self) -> Optional[Iterable[WindowForm]]:
        return

    @property
    def currentForm(self) -> Optional[WindowForm]:
        if not self._currentForm and self.automationInstance:
            top = self.automationInstance.top_window()
            self._currentForm = WindowForm(automationInstance=top)

        return self._currentForm

    @log
    @step
    def focus(self) -> bool:
        if self.automationInstance:
            try:
                self.automationInstance.set_focus()
                return True
            except:
                return False

    @log
    @step
    def start(self) -> ProcessBase:
        super().start()
        app = Application()
        app.start(self._cmdline,
                  work_dir=self._workdir,
                  timeout=self._cs.config.ident.wait_seconds_for_window,
                  retry_interval=self._cs.config.ident.wait_retry_interval_for_window)
        self.automationInstance = app

    @log
    @step
    def snapshot(self) -> Optional[Image]:
        if self.automationInstance:
            return self.automationInstance.capture_as_image()

    def _findForm(self, automationInstance: AutomationInstance) -> Optional[FormBase]:
        for form in self._forms:
            if form.automationInstance == automationInstance:
                return form

    @property
    def parent(self):
        if not self._parent:
            self._parent = Desktop(backend="uia")
        return self._parent



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Iterable, Callable, Any, NewType, Tuple

from .ModelBase import ModelBase

AutomaticInstance = NewType("AutomaticInstance", Any)


@dataclass
class Knowledge(ModelBase):
    alias: Iterable[str]
    launch: Callable[[Knowledge], AutomaticInstance]
    arguments: Tuple[str] = field(default=tuple())



from .TargetBase import TargetBase
from autoinsight.decorator.Step import step
from autoinsight.common.EnumTypes import KnownContextQueries


class SearchBox(TargetBase):
    def __init__(self, query=KnownContextQueries.SearchBox, **kwargs):
        super().__init__(query, **kwargs)
        self._alias.extend(["search", "searchbox", "search_box", "edit"])

    @step
    def type(self, value: str) -> bool:
        if self.automationInstance:
            try:
                if hasattr(self.automationInstance, 'type'):
                    self.automationInstance.type(value)
                else:
                    # https://pywinauto.readthedocs.io/en/latest/code/pywinauto.keyboard.html
                    self.automationInstance.type_keys(value, with_spaces=True)
                return True
            except:
                return False



from autoinsight.script.StepBase import StepBase
from autoinsight.decorator.Log import Log, log


class Step(StepBase):
    def __repr__(self) -> str:
        return f"step {self.name}, action:{self.action}, target:{self._target}, context:{self._context}"

    def __str__(self) -> str:
        return f"step {self.name}({self._target.description}) {self.action} in context: {self._context.description}"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.execute(*args, **kwargs)

    @log
    def execute(self, *args, **kwargs):
        Log.logger.info("Execute %s", self)
        self._isExecuted = True



from time import sleep
from functools import wraps

from autoinsight.decorator.DecoratorBase import DecoratorBase
from autoinsight.services.IoCService import IoCService
from autoinsight.services.ConfigurationServiceBase import ConfigurationServiceBase


class Delay(DecoratorBase):
    _delaySeconds: int = 0

    def __init__(self, seconds: int = 1, *args, **kwargs):
        super().__init__(None, *args, **kwargs)
        self._delaySeconds = seconds

    @classmethod
    @property
    def delaySeconds(cls) -> int:
        if cls._delaySeconds <= 0:
            cs: ConfigurationServiceBase = IoCService().getService(ConfigurationServiceBase)
            cls._delaySeconds = max(cs.config.script.action_delay_seconds, 0)

        return cls._delaySeconds

    def __call__(self, func, *args, **kwargs):
        self._func = func

        if self._delaySeconds > 0:

            @wraps(func)
            def __wrapper(*args, **kwargs):
                sleep(self._delaySeconds)
                return func(*args, **kwargs)

            return __wrapper

        else:
            return func


def delay(func):
    @wraps(func)
    def __wrapper(*args, **kwargs):
        if Delay.delaySeconds > 0:
            sleep(Delay.delaySeconds)
            return func(*args, **kwargs)
        else:
            return func

    return __wrapper



from random import sample
from typing import Callable, Optional, Sequence, Iterable, Tuple
from time import sleep

from .ControlBase import ControlBase
from autoinsight.decorator.Log import log, Log
from autoinsight.decorator.Step import step
from autoinsight.common.Utils import matchScore


class ComboBox(ControlBase):
    def __init__(self, query: str, *args, **kwargs):
        super().__init__(query, *args, **kwargs)
        self._alias.extend(["combobox", "dropdown", "combo box", "drop down"])

    def __repr__(self) -> str:
        pass

    def __str__(self) -> str:
        pass

    @property
    def values(self) -> Sequence[str]:
        if self.isExist():
            try:
                cb = self.automationInstance
                return cb.texts()
            except Exception as e:
                Log.logger.warning("Access values property failed with exception %s", e)
                return False

    @log
    def iterateValues(self, intervalSecond: int = 2,
                      callbackOnEachValue: Optional[Callable[[str], bool]] = None):
        if self.isExist():
            try:
                cb = self.automationInstance
                values = cb.texts()
                for value in values:
                    self.automationInstance.select(value)
                    if callbackOnEachValue:
                        callbackOnEachValue(value)
                    sleep(max(intervalSecond, 1))
                return True
            except Exception as e:
                Log.logger.warning("Iterate values failed with exception %s", e)
                return False

    def _getTexts(self, cb) -> Iterable[str]:
        cb.set_focus()
        cb.expand()
        childList = cb.children(control_type="List")
        if childList:
            return [c.window_text() for c in childList[0].children()]
        else:
            return []

    # TODO: Move the finding logic into the context
    def _getMaxScoreIndex(self, values: Iterable[Tuple], texts: Iterable[str], value: str) -> int:
        if values:
            pairs = [(i, *scores) for i, scores in enumerate(values)]
            score, secondSore = 1, 2
            pairs.sort(key=lambda x: x[score], reverse=True)
            firstScore = pairs[0][score]
            pairs = [c for c in pairs if firstScore == c[score]]
            pairs.sort(key=lambda x: x[secondSore], reverse=True)
            if len(pairs) > 1:
                for pair in pairs:
                    i, score, secondScore = pair
                    if texts[i] == value:
                        return i

            return pairs[0][0]

    @property
    def selected(self):
        if self.isExist():
            return self.automationInstance.selected_text()

    @log
    @step
    def select(self, value: Optional[str] = None, index: int = -1) -> bool:
        if self.isExist():
            try:
                cb = self.automationInstance
                texts = self._getTexts(cb)
                matches = [matchScore(value, [text]) for text in texts]
                if matches:
                    index = self._getMaxScoreIndex(matches, texts, value)
                    text = texts[index]
                    self.automationInstance.select(text)
                    return True
                else:
                    return False
            except Exception as e:
                Log.logger.warning("Select failed with exception %s", e)
                return False

    @log
    @step
    def randomSelect(self) -> bool:
        if self.isExist():
            try:
                cb = self.automationInstance
                values = cb.texts()
                if values:
                    self.automationInstance.select(sample([values], 1)[0])
                    return True
                else:
                    return False
            except Exception as e:
                Log.logger.warning("Random select failed with exception %s", e)
                return False

    @log
    @step
    def selectFirstValue(self) -> bool:
        if self.isExist():
            try:
                cb = self.automationInstance
                values = cb.texts()
                if values:
                    self.automationInstance.select(values[0])
                    return True
                else:
                    return False
            except Exception as e:
                Log.logger.warning("Select first value failed with exception %s", e)
                return False

    @log
    @step
    def selectLastValue(self) -> bool:
        if self.isExist():
            try:
                cb = self.automationInstance
                values = cb.texts()
                if values:
                    self.automationInstance.select(values[-1])
                    return True
                else:
                    return False
            except Exception as e:
                Log.logger.warning("Select last value failed with exception %s", e)
                return False



from time import sleep
from functools import wraps
from typing import Optional

from autoinsight.decorator.DecoratorBase import DecoratorBase
from autoinsight.ident.IdentObjectBase import IdentObjectBase
from autoinsight.services.ContextManagementService import ContextManagementService
from autoinsight.services.IoCService import IoCService
from autoinsight.script.Step import Step as ScriptStep
from autoinsight.script.Script import Script


class Step(DecoratorBase):
    _script: Optional[Script] = None
    """
    This decorator will automation create a step for base on the wrapped function

    The class decorator cannot be used with class member function.
    """

    def __call__(self, *args, **kwargs):
        if self.script and args and isinstance(args[0], IdentObjectBase):
            identObj: IdentObjectBase = args[0]
            s = ScriptStep(type(identObj).__name__,
                           self._func.__name__,
                           identObj,
                           identObj.parent)
            Step.script.addStep(s)
            return self._func(*args, **kwargs)
        else:
            return self._func

    @classmethod
    @property
    def script(cls) -> Optional[Script]:
        if not cls._script:
            cms: ContextManagementService = IoCService().getService(ContextManagementService)
            cls._script = cms.script
        return cls._script


def step(func):
    """
    The decorator should be used with class member functions
    """

    @wraps(func)
    def __wrapper(*args, **kwargs):
        if args and isinstance(args[0], IdentObjectBase):
            identObj: IdentObjectBase = args[0]
            s = ScriptStep(type(identObj).__name__,
                           func.__name__,
                           identObj,
                           identObj.parent)

            if Step.script:
                Step.script.addStep(s)

            return func(*args, **kwargs)
        else:
            return func

    return __wrapper


# TODO: Add class level wrapper
def stepPublic(cls):
    pass



from __future__ import annotations

from abc import abstractmethod
from uuid import UUID

from PIL.Image import Image

from autoinsight.common.ObjectBase import ObjectBase
from autoinsight.common.Utils import GUID


class IdentObjectBase(ObjectBase):
    def __init__(self, *args, **kwargs):
        self._id: UUID = GUID()
        self._repr: str = ""
        self._str: str = ""
        self._automationInstance = None
        self._parent = None

    @property
    def id(self) -> UUID:
        return self._id

    @property
    def description(self) -> str:
        """
        A description that will unique locate the object within the context.

        TODO: The object should be able to rebuild base on the description
        """
        return ""

    @property
    def parent(self) -> IdentObjectBase:
        return self._parent

    @property
    def classname(self):
        """
        The control class name
        """
        return type(self).__name__

    @abstractmethod
    def __repr__(self) -> str:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass

    @abstractmethod
    def snapshot(self) -> Image:
        pass



import sys
from typing import Optional, Dict, Iterable
from pathlib import Path

from .StepBase import StepBase
from autoinsight.common.ObjectBase import ObjectBase
from autoinsight.services.ConfigurationServiceBase import ConfigurationServiceBase
from autoinsight.services.ContextManagementService import ContextManagementService
from autoinsight.services.IoCService import IoCService
from autoinsight.common.CustomTyping import Serializable


class Script(ObjectBase):

    def __init__(self,
                 scriptFile: str = sys.argv[0],
                 ioc: IoCService = IoCService(),
                 runtimeConfig: Optional[Dict[str, Serializable]] = None,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self._ioc: IoCService = ioc
        self._cms: ContextManagementService = ioc.getService(ContextManagementService)
        self._cs: ConfigurationServiceBase = ioc.getService(ConfigurationServiceBase)
        self._runtimeConfig: Optional[Dict[str, Serializable]] = runtimeConfig

        scriptPath = Path(scriptFile)
        self._location: Path = scriptPath.parent
        self._name = scriptPath.stem
        self._cs.updateConfig(self)
        self._steps: list[StepBase] = []

        # TODO changed to event
        self._isPaused = False

    @property
    def name(self) -> str:
        return self._name

    @property
    def location(self) -> str:
        return str(self._location)

    @property
    def steps(self) -> Iterable[StepBase]:
        pass

    def addStep(self, step: StepBase):
        self._steps.append(step)

    @property
    def config(self):
        return self._cs.config

    @property
    def runtimeConfig(self) -> Optional[Dict[str, Serializable]]:
        return self._runtimeConfig

    def run(self):
        for step in self.steps:
            if self._isPaused:
                break

            if not step.isExecuted:
                step.execute(self._isPaused)

    def load(self):
        pass

    def resume(self):
        self._isPaused = False

    def pause(self):
        self._isPaused = True

    def reset(self):
        for step in self.steps:
            step.reset()

    def save(self, path: str) -> bool:
        pass

    def stop(self):
        """
        Will clean and close all the processes start by this script instance
        """
        self.pause()
        self.reset()



from abc import abstractmethod

from .Step import Step


class VerifyBase(Step):
    """
    Verify step for validate a specific condition
    """

    @abstractmethod
    def valid(self, condition, message):
        """
        If the condition is evaluated to false then an exception will be thrown
        """



from .BrowserBase import BrowserBase


class ChromeBrowserBase(BrowserBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def tabs(self):
        pass

    def navigate(self, url: str):
        pass

    def focus(self, window_id: int = None, index: int = None, title: str = None):
        pass

    def close(self):
        pass

    def closeTab(self):
        pass




from typing import Optional, Union

from PIL import Image
from mss.windows import MSS as mss

from .WindowForm import WindowForm
from autoinsight.ident.automation.WindowStartMenuAuto import WindowStartMenuAuto
from autoinsight.ident.automation.WindowSearchBoxAuto import WindowSearchBoxAuto
from autoinsight.ident.target.TargetBase import TargetBase
from autoinsight.common.EnumTypes import KnownContextQueries
from autoinsight.common.Utils import isSimilar, first, matchScore
from autoinsight.common.models.IdentResult import IdentResult
from autoinsight.services.AIServiceBase import AIServiceBase


class WindowStartMenu(WindowForm):
    @property
    def parent(self):
        if not self._parent:
            self._parent = self.desktop
        return self._parent

    @property
    def description(self) -> str:
        return f"the {KnownContextQueries.WindowStartMenu} context"

    def isDraggable(self):
        return False

    def _isQualified(self, result: IdentResult) -> bool:
        return result and result.confidence > 0.6

    def _queryToString(self, query: Union[KnownContextQueries, str]) -> str:
        if isinstance(query, str):
            return query
        else:
            return query.value

    def find(self, query: KnownContextQueries, target: TargetBase) -> Optional[Union[WindowStartMenuAuto, WindowSearchBoxAuto]]:
        with mss() as sct:
            screen = sct.grab(sct.monitors[1])
            screen = Image.frombytes('RGB', screen.size, screen.bgra, 'raw', 'BGRX')
            ident: AIServiceBase = self._ioc.getService(AIServiceBase)
            results: IdentResult = ident.predict(screen)

            # 0: button_create 1: search_box 2: start_bar
            if query == KnownContextQueries.WindowStartMenu:
                result: IdentResult = first(results, isRevert=True, filterFunc=lambda x: x.classnum == 2, sortKeyFunc=lambda x: x.confidence)
                if result and self._isQualified(result):
                    return WindowStartMenuAuto(self._ioc, result.rect)
            elif isSimilar(query, KnownContextQueries.SearchBox):
                result: IdentResult = first(results, isRevert=True, filterFunc=lambda x: x.classnum == 1, sortKeyFunc=lambda x: x.confidence)
                if result and self._isQualified(result):
                    return WindowSearchBoxAuto(self._ioc, result.rect)
                else:
                    windows = self.desktop.windows()
                    window = first(filter(lambda w: w.window_text() == "Start" and w.children() and w.children()[0].window_text() == "Search", windows))
                    self._automationInstance = window
                    return super().find(self._queryToString(query), target)





from typing import Iterable, Optional

from PIL.Image import Image

from .FormBase import FormBase
from .WindowGUIContextBase import WindowGUIContextBase
from .ProcessBase import ProcessBase
from autoinsight.ident.target.TargetBase import TargetBase
from autoinsight.decorator.Log import log
from autoinsight.decorator.Step import step
from autoinsight.common.EnumTypes import ScrollAmountTypes, ScrollDirectionTypes


class WindowForm(FormBase, WindowGUIContextBase):

    def __int__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._title = ""

    @property
    def description(self) -> str:
        return f"Window form {self._title}"

    def __repr__(self):
        pass

    def __str__(self):
        pass

    @property
    def targets(self) -> Iterable[TargetBase]:
        pass

    @property
    def title(self) -> str:
        if self.automationInstance:
            return self.automationInstance.window_text()

    @log
    @step
    def snapshot(self) -> Optional[Image]:
        if self.automationInstance:
            try:
                return self.automationInstance.capture_as_image()
            except:
                return

    @log
    @step
    def scroll(self,
               direction: ScrollDirectionTypes = ScrollDirectionTypes.Down,
               amount: ScrollAmountTypes = ScrollAmountTypes.Lines,
               count: int = 1):

        if self.automationInstance:
            return self.automationInstance.scroll(direction, amount, count)

    @log
    @step
    def focus(self):
        if self.automationInstance:
            self.automationInstance.set_focus()

    @log
    @step
    def maximize(self):
        if self.automationInstance:
            self.automationInstance.maximize()

    @log
    @step
    def minimize(self):
        if self.automationInstance:
            self.automationInstance.minimize()

    @log
    @step
    def drag(self):
        pass

    @log
    @step
    def drop(self):
        pass

    @log
    @step
    def dragTo(self, x: int, y: int):
        pass

    @log
    @step
    def start(self) -> ProcessBase:
        """
        For will do noting for the start method
        """
        self.focus()



from os.path import isfile
from typing import Optional, Sequence

import cv2 as cv
import colour
import numpy as np
from PIL.Image import Image, open as imageopen, fromarray

from .ImageProcessingServiceBase import ImageProcessingServiceBase
from autoinsight.common.models.Rectangle import Rectangle
from autoinsight.decorator.Log import log
from autoinsight.common.models.Size import Size
from autoinsight.common.EnumTypes import CVImageTypes
from autoinsight.common.CustomTyping import IMG


class LocalImageProcessingService(ImageProcessingServiceBase):
    @log
    def getImageResolution(self, image: IMG) -> Optional[Size]:
        image = self._img2pil(image)
        result = Size(*image.size)
        image.close()
        return result

    @log
    def compareBrightness(self, imageA: IMG, imageB: IMG) -> int:
        """
        Compare two image brightness

        @param imageA PIL image
        @param imageB PIL image
        @return negative, 0, positive integer
        """
        imageA = self._img2pil(imageA)
        imageB = self._img2pil(imageB)
        cvImageA = self._pil2cv(imageA)
        cvImageB = self._pil2cv(imageB)
        histA = cv.calcHist([self._toGray(cvImageA)], [0], None, [256], [0, 256])
        histB = cv.calcHist([self._toGray(cvImageB)], [0], None, [256], [0, 256])
        result = np.sum((histA - histB) * np.array([[x] for x in range(0, 256)]))
        imageA.close()
        imageB.close()
        return int(result)

    @log
    def compareSharpness(self, imageA: IMG, imageB: IMG) -> int:
        pass

    @log
    def compareNoise(self, imageA: IMG, imageB: IMG) -> int:
        pass

    @log
    def compareFocusResult(self, imageA: IMG, imageB: IMG, rectangle: Rectangle) -> int:
        pass

    @log
    def compareImageISO(self, imageA: IMG, imageB: IMG) -> int:
        pass

    @log
    def compareVideoISO(self, videoA: Image, videoB: Image, checkpoint: float = 0.0) -> int:
        pass

    @log
    def compareImageContrast(self, imageA: IMG, imageB: IMG) -> int:
        imageA = self._img2pil(imageA)
        imageB = self._img2pil(imageB)
        cvImageA = self._toGray(self._pil2cv(imageA))
        cvImageB = self._toGray(self._pil2cv(imageB))
        histA = cv.calcHist([cvImageA], [0], None, [256], [0, 256])
        histB = cv.calcHist([cvImageB], [0], None, [256], [0, 256])
        contrastMeasureA = self._measureContrast(histA.flatten())
        contrastMeasureB = self._measureContrast(histB.flatten())
        return contrastMeasureA - contrastMeasureB

    @log
    def compareVideoContrast(self, videoA, videoB, checkpoint: float = 0.0) -> int:
        pass

    @log
    def compareImageGamma(self, imageA: IMG, imageB: IMG) -> int:
        """
        TODO: Add gamma calculation logic
        """
        return self.compareBrightness(imageA, imageB)

    @log
    def compareVideoGamma(self, videoA, videoB, checkpoint: float = 0.0) -> int:
        pass

    @log
    def getBSODInfo(self, image) -> str:
        pass

    @log
    def getImageColorTemperature(self, image: IMG) -> float:
        r, g, b = self.getImageRGBMean(image)
        # Assuming sRGB encoded colour values.
        RGB = np.array([r, g, b])
        # Conversion to tristimulus values.
        XYZ = colour.sRGB_to_XYZ(RGB / 255)
        # Conversion to chromaticity coordinates.
        xy = colour.XYZ_to_xy(XYZ)
        # Conversion to correlated colour temperature in K.
        return colour.xy_to_CCT(xy, 'hernandez1999')

    @log
    def getVideoColorTemperature(self, video, checkpoint: float = 0.0) -> int:
        pass

    @log
    def getImageHSVMean(self, image: IMG):
        image = self._img2pil(image)
        cvimage = self._pil2cv(image, CVImageTypes.BGR)
        hsv = cv.cvtColor(cvimage, cv.COLOR_BGR2HSV)
        image.close()
        return np.mean(hsv, axis=(0, 1))

    @log
    def getVideoHSVMean(self, video, checkpoint: float = 0.0) -> int:
        pass

    @log
    def getImageRGBMean(self, image: IMG) -> Sequence[int]:
        image = self._img2pil(image)
        cvimage = self._pil2cv(image, CVImageTypes.RGB)
        image.close()
        return np.mean(cvimage, axis=(0, 1))

    @log
    def getVideoRGBMean(self, video, checkpoint: float) -> Sequence[int]:
        pass

    @log
    def getBrightestRegion(self, image: IMG) -> Rectangle:
        pass

    @log
    def getDarkestRegion(self, image: IMG) -> Rectangle:
        pass

    @log
    def getImageISO(self, image: IMG) -> int:
        pass

    @log
    def getVideoISO(self, video, checkpoint: float = 0.0) -> int:
        pass

    @log
    def isVideoStutter(self, video) -> bool:
        pass

    @log
    def isVideoFlicker(self, video) -> bool:
        pass

    @log
    def isContainObject(self, image: IMG, target: IMG) -> bool:
        image = self._img2pil(image)
        target = self._img2pil(target)
        cvimage = self._toGray(self._pil2cv(image))
        cvtarget = self._toGray(self._pil2cv(target))
        image.close()
        target.close()
        return np.any(np.where(cv.matchTemplate(cvimage, cvtarget, cv.TM_CCOEFF_NORMED) > 0.4))

    @log
    def isPrivacyLEDOn(self, imageA: IMG, imageB: IMG) -> bool:
        pass

    @log
    def isImageCorrupted(self, image: IMG) -> bool:
        pass

    @classmethod
    @log
    def _img2pil(cls, image: IMG) -> Image:
        if isinstance(image, str) and isfile(image):
            return imageopen(image)
        elif isinstance(image, np.ndarray):
            return fromarray(image)
        else:
            return image

    @classmethod
    @log
    def _toGray(cls, image: np.ndarray):
        if len(image.shape) > 2:
            image = cv.cvtColor(image, cv.COLOR_RGB2GRAY)
        return image

    @classmethod
    @log
    def _pil2np(cls, image: IMG):
        return np.frombuffer(image.tobytes(), dtype=np.uint8)

    @classmethod
    @log
    def _pil2cv(cls, image: IMG, imageType: CVImageTypes = CVImageTypes.RGB):
        image = cls._img2pil(image)

        if imageType == CVImageTypes.RGB:
            nparray = cls._pil2np(image.convert('RGB'))
            # 3 or 4 base on your image mode is RGB or RGBA
            nparray.shape = (*reversed(image.size), 3)
            return nparray
        elif imageType == CVImageTypes.RGBA:
            nparray = cls._pil2np(image.convert('RGBA'))
            nparray.shape = (*reversed(image.size), 4)
            return nparray
        else:
            nparray = cls._pil2np(image.convert('RGB'))
            nparray.shape = (*reversed(image.size), 3)
            return cv.cvtColor(nparray, cv.COLOR_RGB2BGR)

    @classmethod
    @log
    def _measureContrast(cls, histogram: np.ndarray) -> int:
        leftIndex = next(i for i in range(len(histogram)) if histogram[i] != 0)
        rightIndex = next(i for i in range(len(histogram) - 1, -1, -1) if histogram[i] != 0)
        points = []
        i, j, leftValue, rightValue, previous = 0, 0, 0, 0, 0

        while True:
            left = leftIndex + i
            right = rightIndex - j
            leftValue -= histogram[left]
            rightValue += histogram[right]
            total = leftValue + rightValue

            if total < 0:
                leftValue += rightValue
                j += 1
            elif total > 0:
                rightValue += leftValue
                i += 1
            else:
                i += 1
                j += 1

            value = min(abs(leftValue), abs(rightValue)) * 2

            if total * previous <= 0:
                points.append(value)
            else:
                points[-1] += value

            if left >= right:
                break
            else:
                previous = total

        points.reverse()
        return int(sum(((i + 1) * v for i, v in enumerate(points))))



import os.path
import time
from functools import wraps

from autoinsight.decorator.DecoratorBase import DecoratorBase
from autoinsight.modules.Logging import ConsoleAndFileLogger, LoggerBase, ConsoleLogger
from autoinsight.services.IoCService import IoCService
from autoinsight.services.ConfigurationServiceBase import ConfigurationServiceBase


class Log(DecoratorBase):
    """
    function log decorator
    """
    _logger: LoggerBase = None

    @classmethod
    @property
    def logger(cls) -> LoggerBase:
        if not cls._logger:
            cs: ConfigurationServiceBase = IoCService().getService(ConfigurationServiceBase)

            if cs.config.script and cs.config.script.output_root:
                logPath = os.path.join(cs.config.script.output_root,
                                       time.strftime(f"{cs.scriptName}-%Y%m%d%H%M%S",
                                                     time.gmtime()))
                cls._logger = ConsoleAndFileLogger(logPath,
                                                   "all",
                                                   cs.config.common.log_level,
                                                   cs.config.common.log_format)
            else:
                cls._logger = ConsoleLogger("all",
                                            cs.config.common.log_level,
                                            cs.config.common.log_format)
        return cls._logger

    def __call__(self, *args, **kwargs):
        try:
            func = self._func
            self.logger.debug("Call %s.%s with %s, %s",
                              self._func.__module__,
                              self._func.__name__,
                              args,
                              kwargs)
            ret = func(*args, **kwargs)
            self.logger.debug("%s.%s return %s",
                              self._func.__module__,
                              self._func.__name__,
                              ret)
            return ret
        except Exception as e:
            self.logger.error(f"Call function {self._func.__module__}.{self._func.__name__} with error {e}")
            raise e


def log(func):
    """
    logging decorator to a function and member function. When it is used with classmethod
    please add this decorator before the classmethod decorator.
    """
    @wraps(func)
    def __wrapper(*args, **kwargs):
        try:
            moduleName = func.__module__ if hasattr(func, "__module__") else ""
            funcName = func.__name__ if hasattr(func, "__name__") else ""
            Log.logger.debug("Call %s.%s with %s, %s",
                             moduleName,
                             funcName,
                             args,
                             kwargs)
            ret = func(*args, **kwargs)
            Log.logger.debug("%s.%s return %s",
                             moduleName,
                             funcName,
                             ret)
            return ret
        except Exception as e:
            Log.logger.error(f"Call function {func.__module__}.{func.__name__} with error {e}")
            raise e

    return __wrapper


# TODO: Add class level log wrapper
def logPublic(cls):
    pass



from autoinsight.common.ObjectBase import ObjectBase


class DriverBase(ObjectBase):
    def __init__(self, *args, **kwargs):
        pass

    def wait(self, seconds: int):
        pass

    def find(self, *args, **kwargs):
        pass




import dataclasses
from typing import Iterable, Any, Dict, Tuple

from autoinsight.common.ObjectBase import ObjectBase


@dataclasses.dataclass
class ModelBase(ObjectBase):
    """
    All the subclass inherit from ModelBase should be decorated with dataclass
    """
    def __iter__(self):
        pass

    def asTuple(self) -> Tuple:
        return dataclasses.astuple(self)

    def asDict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)

    @property
    def values(self) -> Iterable[Any]:
        return (getattr(self, field.name) for field in self.fields)

    @property
    def fields(self) -> Iterable[dataclasses.Field]:
        return dataclasses.fields(self)

    @property
    def keys(self) -> Iterable[str]:
        return (field.name for field in self.fields)



from dataclasses import dataclass

from .ModelBase import ModelBase
from .Point import Point
from .Size import Size


@dataclass
class ShapeBase(ModelBase):
    left: int
    top: int

    @property
    def center(self) -> Point:
        pass

    @property
    def size(self) -> Size:
        pass



from abc import ABC
from time import sleep
from typing import Optional, Iterable, Any, Mapping, Tuple, Union

from pywinauto import Desktop, Application, WindowSpecification
from pywinauto.controls.uiawrapper import UIAWrapper
from pywinauto.findbestmatch import MatchError

from autoinsight.common.CustomTyping import AutomationInstance, ElementsInfo, FoundCtrlItem
from autoinsight.common.Utils import matchScore, isIEqual, isSimilar, first, similarRate
from autoinsight.decorator.Log import log, Log
from autoinsight.ident.target.TargetBase import TargetBase
from autoinsight.ident.context.GUIContextBase import GUIContextBase


class WindowGUIContextBase(GUIContextBase, ABC):
    _desktop = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _isContainerTypes(self, classname: str) -> bool:
        return any(similarRate(classname, c) > 0.5 for c in self.containers)

    def _isSimilarTypes(self, classname, types: Iterable[str]) -> bool:
        return any(similarRate(classname, t) > 0.5 for t in types)

    def _getAllCtrlMapFromElementsInfo(self, info: ElementsInfo, target: TargetBase = None) -> Iterable[Tuple[Iterable[Any], Mapping[int, Iterable[str]]]]:
        allCtrlMapPairs = []
        children = []
        ctrlInfo = info
        while True:
            if target:
                invalidIndexes = (i for i, c in enumerate(ctrlInfo.allCtrl) if not (self._isSimilarTypes(c.friendly_class_name(), target.aliases)
                                                                                    or self._isContainerTypes(c.friendly_class_name())))
                for index in invalidIndexes:
                    if index in ctrlInfo.allCtrlIndexNameMaps:
                        del ctrlInfo.allCtrlIndexNameMaps[index]

            allCtrlMapPairs.append((ctrlInfo.allCtrl, ctrlInfo.allCtrlIndexNameMaps))
            for index in ctrlInfo.allCtrlIndexNameMaps.keys():
                ctrl = ctrlInfo.allCtrl[index]
                if hasattr(ctrl, 'allCtrl') and ctrl.allCtrl:
                    children.append(ctrl)

            if children:
                ctrlInfo = children.pop()
            else:
                break
        return allCtrlMapPairs

    def _getTexts(self, ctrl: UIAWrapper) -> str:
        texts = []
        if ctrl:
            try:
                texts.append(ctrl.window_text())
                texts.append(ctrl.friendly_class_name())
            except:
                pass
        return texts

    def _getAllCtrlMapFromUIAWrapper(self, info: UIAWrapper, target: TargetBase = None) -> Iterable[Tuple[Iterable[Any], Mapping[int, Iterable[str]]]]:
        allCtrlMapPairs = []
        children = []
        ctrlInfo = info

        while True:
            infoChildren = ctrlInfo.children()
            infoCtrlIndexNameMaps = {k: self._getTexts(v) for k, v in enumerate(infoChildren)}

            if target:
                invalidIndexes = [i for i, c in enumerate(infoChildren) if not (self._isSimilarTypes(c.friendly_class_name(), target.aliases)
                                                                                or self._isContainerTypes(c.friendly_class_name()))]
                for index in invalidIndexes:
                    del infoCtrlIndexNameMaps[index]

            allCtrlMapPairs.append((infoChildren, infoCtrlIndexNameMaps))
            for index in infoCtrlIndexNameMaps.keys():
                ctrl = infoChildren[index]
                children.append(ctrl)

            if children:
                ctrlInfo = children.pop()
            else:
                break

        return allCtrlMapPairs

    def _getAllCtrlMap(self, info: Union[ElementsInfo, UIAWrapper], target: TargetBase = None) -> Iterable[Tuple[Iterable[Any], Mapping[int, Any]]]:
        if isinstance(info, ElementsInfo):
            return self._getAllCtrlMapFromElementsInfo(info, target)
        elif isinstance(info, UIAWrapper):
            return self._getAllCtrlMapFromUIAWrapper(info, target)

    def _getMaxScoreItem(self, controls: Iterable[FoundCtrlItem], target: TargetBase) -> Optional[Any]:
        if controls:
            firstIndex, score, secondSore = 0, 1, 2
            controls.sort(key=lambda x: x[score], reverse=True)
            controls = [c for c in controls if controls[firstIndex][score] == c[score]]
            controls.sort(key=lambda x: x[secondSore], reverse=True)

            if len(controls) > 1 and target:
                firstCandidate = controls[firstIndex]
                candidates = [x for x in filter(lambda x: firstCandidate[score] == x[score] and firstCandidate[secondSore] == x[secondSore], controls)]
                return first(candidates, isRevert=True, sortKeyFunc=lambda x: similarRate(x.ctrl.friendly_class_name(), target.classname)).ctrl

            return controls[firstIndex].ctrl

    def _getClickable(self, ctrl: Any, foundIndex: int) -> Any:
        if ctrl and isIEqual(ctrl.friendly_class_name(), 'Hyperlink'):
            index = -1
            controls = ctrl.parent().children()
            for i, c in enumerate(controls):
                if c == ctrl:
                    index = i
                    break

            for i in range(index - 1, -1, -1):
                ctrl = ctrl.parent().children()[i]
                if isIEqual(ctrl.friendly_class_name(), 'Button'):
                    return ctrl

        return ctrl

    def _getElementInfo(self, context: Any, query, target: TargetBase) -> ElementsInfo:
        try:
            if isinstance(context, Application):
                context = context.top_window()
            elif isinstance(context, UIAWrapper):
                return context

            return context.get_elements_info(depth=self._searchDepth, max_width=self._searchWidth)
        except MatchError:
            return self._findFromDesktop(query, target, context)

    def _findFromDesktop(self, query: str, target: TargetBase, context: ElementsInfo) -> Any:
        window: WindowSpecification = None
        for w in self.desktop.windows():
            if w.process_id() == self.processId and w.is_active() and not isSimilar(w.friendly_class_name(), 'tooltip'):
                window = w
                break

        if not window:
            window = WindowGUIContextBase.desktop.top_from_point(self.rectangle.center.x, self.rectangle.center.y)

        if window:
            self._automationInstance = window
            window.wait_for_idle()
            return window

    @property
    def parent(self) -> Any:
        if not self._parent:
            self._parent = self.desktop
        return self._parent

    @classmethod
    @property
    def desktop(cls) -> Desktop:
        if not cls._desktop:
            cls._desktop = Desktop(backend="uia")
        return cls._desktop

    def _toStr(self, value):
        try:
            return str(value)
        except:
            return ""

    @log
    def find(self, query: str, target: TargetBase = None, *args, **kwargs) -> Optional[AutomationInstance]:
        if self.automationInstance:
            try:
                info: ElementsInfo = self._getElementInfo(self.automationInstance, query, target)
                allCtrlMapPairs = self._getAllCtrlMap(info, target)
                foundControls = []
                for pair in allCtrlMapPairs:
                    bestIndex, bestScore, bestSecond = -1, -1, -1
                    allCtrl, allCtrlIndexMaps = pair
                    otherCandidateIndexes = []
                    for index, names in allCtrlIndexMaps.items():
                        score, secondScore = matchScore(query, names)
                        if score > bestScore:
                            bestIndex = index
                            bestScore = score
                            bestSecond = secondScore
                            otherCandidateIndexes.clear()

                        elif score == bestScore and secondScore > bestSecond:
                            bestIndex = index
                            bestScore = score
                            bestSecond = secondScore
                            otherCandidateIndexes.clear()

                        elif score > 0 and secondScore > 0 and score == bestScore and secondScore == bestSecond:
                            otherCandidateIndexes.append(index)

                    if bestIndex > -1 and bestScore > 0 and bestSecond > 0:
                        foundControls.append(FoundCtrlItem(allCtrl[bestIndex], bestScore, bestSecond, bestIndex))

                        if otherCandidateIndexes:
                            for index in otherCandidateIndexes:
                                foundControls.append(FoundCtrlItem(allCtrl[index], bestScore, bestSecond, index))
                            otherCandidateIndexes.clear()

                if foundControls:
                    ctrl = self._getMaxScoreItem(foundControls, target)
                    return ctrl
                else:
                    Log.logger.warning(f"Find target {self._toStr(target)} failed with {query} in context {self._toStr(self.automationInstance)}")

            except Exception as e:
                Log.logger.warning("Find %s from context failed with error %s", query, e)
                return



import setuptools

from os.path import dirname, join


current_path = dirname(__file__)

with open(join(current_path, "README.md"), "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open(join(current_path, "requirements.txt"), "r", encoding="utf=8") as fh:
    requires = fh.readlines()


def get_version(rel_path):
    with open(rel_path) as fh:
        for line in fh.readlines():
            if line.startswith('__version__'):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
        else:
            raise RuntimeError("Unable to find version string.")


setuptools.setup(
    name="autoinsight",
    version=get_version("src/autoinsight/__init__.py"),
    author="Erxin(Edwin) Shang",
    author_email="author@example.com",
    description="A Simplified UI automation package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/shangerxin/autoinsight",
    project_urls={
        "Bug Tracker": "https://github.com/shangerxin/autoinsight/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: Microsoft :: Windows",
    ],
    install_requires=requires,
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    include_package_data=True,
    package_data={
        "": ["*.txt", "*.yml", "*.json", "*.pt"]
    },
    data_files=["requirements.txt"]
)



import os
import sys
from typing import Iterable

from ultralytics import YOLO

from .AIServiceBase import AIServiceBase
from autoinsight.common.models.Rectangle import Rectangle
from autoinsight.common.models.IdentResult import IdentResult


class YoloAIService(AIServiceBase):
    def __init__(self, *args, **kwargs):
        projectDir = os.path.normpath(__file__).replace(r"src\autoinsight\services\YoloAIService.py", "")
        self._model = YOLO(os.path.join(projectDir, r"src\autoinsight\ident\models\yolov8\yolov8l_custom.pt"))

    # 0: button_create   1: search_box   2: start_bar
    def predict(self, image) -> Iterable[IdentResult]:
        results = self._model.predict(source=image, show=False, save=True, save_txt=True, line_width=1, save_conf=True, conf=0.6)

        identResults = []
        for result in results:
            result = result.numpy()
            boxes = result.boxes
            for i, box in enumerate(boxes.xywh):
                confidence = boxes.conf[i]
                classnum = boxes.cls[i]
                left, top, width, height = (int(x) for x in box)
                identResults.append(IdentResult(Rectangle(left, top, width, height), confidence, classnum))

        return identResults



from __future__ import annotations

import os
import signal
from abc import abstractmethod
from typing import Optional
from subprocess import Popen

from autoinsight.common.CustomTyping import AutomationInstance
from autoinsight.decorator.Log import log
from autoinsight.ident.IdentObjectBase import IdentObjectBase


class ProcessBase(IdentObjectBase):
    def __init__(self,
                 processId: Optional[int] = 0,
                 processName: Optional[str] = None,
                 processHandle: Optional[Popen] = None,
                 cmdline: Optional[str] = None,
                 title: Optional[str] = None,
                 workdir: Optional[str] = None,
                 automationInstance: Optional[AutomationInstance] = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._processId: int = processId
        self._processHandle: Optional[Popen] = processHandle
        self._processName: str = processName
        self._cmdline: str = cmdline
        self._title: str = title
        self._workdir: str = workdir
        self._exitcode: int = 0

        if automationInstance:
            self.automationInstance: Optional[AutomationInstance] = automationInstance
            self._isStarted = True
        else:
            self._automationInstance: Optional[AutomationInstance] = None
            self._isStarted: bool = False

    def __repr__(self):
        if not self._repr:
            self._repr = f"name:{self._processName} id:{self.processId} cmdline:{self.cmdline} workdir:{self.workdir}"
        return self._repr

    def __str__(self):
        if not self._str:
            if self.cmdline:
                self._str = os.path.basename(self.cmdline)
            elif self._title:
                self._str = self._title
            elif self._processName:
                self._str = self._processName
            elif self._processId:
                self._str = str(self._processId)
            else:
                self._str = ""

        return self._str

    @property
    def parent(self) -> IdentObjectBase:
        return self._parent

    @property
    def automationInstance(self) -> Optional[AutomationInstance]:
        if not self._isStarted:
            self.start()

        if not self._automationInstance:
            self.automationInstance = self._cms.os.find(
                f"{self._processId} {self.cmdline} {self._title} {self.workdir}")
        return self._automationInstance

    @automationInstance.setter
    def automationInstance(self, value: AutomationInstance):
        if value and value != self._automationInstance:
            self._automationInstance = value
            if hasattr(value, 'process') and isinstance(value.process, int):
                self._processId = value.process
            elif hasattr(value, 'process_id'):
                self._processId = value.process_id()

    @property
    def exitcode(self) -> int:
        return self._exitcode

    @property
    def workdir(self) -> str:
        return self._workdir

    @property
    def cmdline(self) -> str:
        return self._cmdline

    @property
    def processName(self) -> str:
        return self._processName

    @property
    def processId(self) -> int:
        return self._processId

    def __enter__(self):
        super().__enter__()
        self.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def start(self) -> ProcessBase:
        self._isStarted = True

    @classmethod
    @abstractmethod
    def new(cls, *args, **kwargs) -> ProcessBase:
        pass

    @log
    def close(self):
        if self._processId:
            os.kill(self._processId, signal.SIGTERM)
            self.tearDown()
            return

        if self._processHandle:
            self._processHandle.terminate()
            return

        if self.automationInstance:
            self.automationInstance.close()
            self.tearDown()
            return

    @log
    def kill(self):
        if self._processId:
            os.kill(self._processId, signal.SIGKILL)
            self.tearDown()
            return

    def _printElementsTree(self):
        """
        Print debug element tree information to the console.
        """
        pass



from .ProcessBase import ProcessBase
from .ShellBase import ShellBase


class WindowShellBase(ShellBase):
    def start(self) -> ProcessBase:
        pass

    @classmethod
    def new(cls, *args, **kwargs) -> ProcessBase:
        pass

    def snapshot(self):
        pass

    def __init__(self, *args, **kwargs):
        self.args = args
        super().__init__(*args, **kwargs)



from PIL.Image import Image

from autoinsight.ident.context.WindowGUIContextBase import WindowGUIContextBase

class WindowDesktop(WindowGUIContextBase):
    @classmethod
    def new(cls, *args, **kwargs) -> ProcessBase:
        pass

    def switch(self):
        pass

    def snapshot(self) -> Image:
        pass



from enum import Enum, auto

import autoinsight.extend.ExtendEnum  # noqa:F401


class OSTypes(Enum):
    Any = "Any"
    Windows = "Windows"
    Linux = "Linux"
    Java = "Java"
    MacOS = "Darwin"


class ButtonTypes(Enum):
    Left = "left"
    Middle = "middle"
    Right = "right"


class ScrollDirectionTypes(Enum):
    Up = "up"
    Down = "down"
    Left = "left"
    Right = "right"


class ScrollAmountTypes(Enum):
    Lines = "lines"
    Page = "page"
    End = "end"


class CVImageTypes(Enum):
    RGBA = "RGBA"
    RGB = "RGB"
    BGR = "BGR"


class EventTypes(Enum):
    pass


class GUIContainerTypes(Enum):
    Dialog = "Dialog"
    Panel = "Panel"
    Form = "Form"
    Window = "Window"
    Pane = "Pane"
    Group = "Group"


class BinaryTypes(Enum):
    Bit16 = auto()
    Bit32 = auto()
    Bit64 = auto()


class KnownContextQueries(Enum):
    WindowStartMenu = "Window start menu"
    SearchBox = "Search box"
